// KeyboardDialog.cpp : implementation file
//
#include "stdafx.h"
#include "resource.h"
#include "KeyboardDialog.h"
#include "mikmod.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include "globals.h"

#define INTERVAL 1.0594631			//in well-tempered scale, a pitch x
									//which is n half-steps from pitch y
									//is ~ x = y + n ^ INTERVAL

SAMPLE *psample;

//Used to track last note when dragging over keys.
//Since we play notes on MouseMove as well as 
//on LButtonDown, we need to prevent the same key
//being hit a number of times by dragging over it.
//So when dragging over keys we only trigger a 
//PlayNote when key changes.
static int nLastNotePlayed = -1;

void CALLBACK EXPORT KeyboardTimerProc(
   HWND hWnd,      // handle of CWnd that called SetTimer
   UINT nMsg,      // WM_TIMER
   UINT nIDEvent,   // timer identification
   DWORD dwTime    // system time
)
{
	MikMod_Update();
}

void PlayNote(int nOffset)
{
	ULONG freq;
	int voice;

	// nOffset is offset from sample's pitch in half-steps.
	// I have no idea if that's supposed to be a middle C or
	// what; MikMod reports the frequency of most samples as
	// 8363 Hz.  All I do here is take the sample's pitch and
	// modify it by nOffset half-steps (well-tempered scale).

	freq = (ULONG)(psample->speed * pow(INTERVAL, nOffset));

	voice = Sample_Play(psample,0,0);
	Voice_SetFrequency(voice, freq);	
}


/////////////////////////////////////////////////////////////////////////////
// CKeyboardDialog dialog


CKeyboardDialog::CKeyboardDialog(CWnd* pParent /*=NULL*/)
	: CDialog(CKeyboardDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(CKeyboardDialog)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CKeyboardDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CKeyboardDialog)
	DDX_Control(pDX, IDC_SAMPLE_LIST, m_SampleList);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CKeyboardDialog, CDialog)
	//{{AFX_MSG_MAP(CKeyboardDialog)
	ON_NOTIFY(NM_CLICK, IDC_SAMPLE_LIST, OnClickSampleList)
	ON_WM_LBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_BN_CLICKED(IDC_STOPIT, OnStopIt)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CKeyboardDialog message handlers

BOOL CKeyboardDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	CString cstrSampleName;
	int iSample;

	// Samples don't have names we can get in MikMod, although
	// "instruments" do -- but not all modules support instruments.
	// So this is just a simple numbered list of samples.
	for(iSample=0; iSample < m_module->numsmp; iSample++) {
		cstrSampleName.Format(_T("S%d"), iSample);
		m_SampleList.InsertItem(LVIF_TEXT,iSample,cstrSampleName,0,0,0,iSample);
	}

	// set small column width
	m_SampleList.SetColumnWidth(-1, 5);

    MikMod_SetNumVoices(-1, 32);
    MikMod_EnableOutput();
	
	//select first sample as default
	psample = &(m_module->samples[0]);

	//setup timer
	SetTimer(ID_KEYBOARD_TIMER, (nKeyboardTimerFrequency * nKeyboardBufferSize) / 100, &KeyboardTimerProc);
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CKeyboardDialog::OnClickSampleList(NMHDR* pNMHDR, LRESULT* pResult) 
{
	POSITION pos;
	int item;
	CString strSampleInfo;
	
	pos = m_SampleList.GetFirstSelectedItemPosition();
	item = m_SampleList.GetNextSelectedItem(pos);
	
	psample = &(m_module->samples[item]);

	Sample_Play(psample,0,0);
	
	strSampleInfo.Format(_T("Length: %d samples.  Loops: %c"), 
		psample->length, (psample->flags & SF_LOOP) ? 'Y' : 'N');
	this->SetDlgItemText(IDC_SAMPLEINFO_STATIC, strSampleInfo);

	*pResult = 0;
}


void CKeyboardDialog::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int nNote;

	nNote = GetKeyPlayed(point);
	if (nNote >= 0)
		PlayNote(nNote);
	
	CDialog::OnLButtonDown(nFlags, point);
}

void CKeyboardDialog::OnMouseMove(UINT nFlags, CPoint point) 
{
	int nNote;

	nNote = GetKeyPlayed(point);
	if ((nNote >= 0) && (nNote != nLastNotePlayed)) {
		PlayNote(nNote);
		nLastNotePlayed = nNote;
	}
	
	CDialog::OnMouseMove(nFlags, point);
}

int CKeyboardDialog::GetKeyPlayed(CPoint point)
{
	int nNote;
	int nWhiteKey;
	bool bBlack = false;
	bool bWhite = false;

	//FIXME: arbitrary numbers, no system metrics, blah blah blah

	int nYOffset = 2;
	int nYKeySpacing = 34;
	int nXBlacksStart = 80;
	int nXKeysEnd = 132;
	int nYHalfBlack = 9;

	//calc note based on position in keyboard

	//if in black keys area
	if ((point.x < nXKeysEnd) && (point.x > nXBlacksStart)) {

		//Tortuous and not very clean testing. If we wanted to be
		//correct, we'd draw user-selectable and systemmetrics-compatible
		//keyboards, and do neat calculations.

		//first black is at nYOffset plus nYKeySpacing plus or minus nYHalfBlack,
		//rest are at irregular nYKeySpacing intervals
		if ( (point.y > nYOffset + nYKeySpacing - nYHalfBlack) &&
			 (point.y < nYOffset + nYKeySpacing + nYHalfBlack) ) {
			nNote = 1;
			bBlack = true;
		}
		else if ( (point.y > nYOffset + 2 * nYKeySpacing - nYHalfBlack) &&
			 (point.y < nYOffset + 2 * nYKeySpacing + nYHalfBlack) ) {
			nNote = 3;
			bBlack = true;
		}
		else if ( (point.y > nYOffset + 4 * nYKeySpacing - nYHalfBlack) &&
			 (point.y < nYOffset + 4 * nYKeySpacing + nYHalfBlack) ) {
			nNote = 6;
			bBlack = true;
		}
		else if ( (point.y > nYOffset + 5 * nYKeySpacing - nYHalfBlack) &&
			 (point.y < nYOffset + 5 * nYKeySpacing + nYHalfBlack) ) {
			nNote = 8;
			bBlack = true;
		}
		else if ( (point.y > nYOffset + 6 * nYKeySpacing - nYHalfBlack) &&
			 (point.y < nYOffset + 6 * nYKeySpacing + nYHalfBlack) ) {
			nNote = 10;
			bBlack = true;
		}

		//if in black area but no black hit, assume white hit
		if (!bBlack) {
			nWhiteKey = (point.y - nYOffset) / nYKeySpacing;
			bWhite = true;
		}
	}
	//if in white key area
	else if (point.x < nXBlacksStart) {
		nWhiteKey = (point.y - nYOffset) / nYKeySpacing;
		bWhite = true;
	}

	//select note for white key hit
	if (bWhite)
		switch (nWhiteKey) {
		case 0:
			nNote = 0;
			break;
		case 1:
			nNote = 2;
			break;
		case 2:
			nNote = 4;
			break;
		case 3:
			nNote = 5;
			break;
		case 4:
			nNote = 7;
			break;
		case 5:
			nNote = 9;
			break;
		case 6:
			nNote = 11;
			break;
		default:
			nNote = 0;
		}

	if (bWhite || bBlack) 
		return nNote;
	else
		return -1;
}

void CKeyboardDialog::OnStopIt() 
{
    MikMod_DisableOutput();
    MikMod_SetNumVoices(-1, 32);
    MikMod_EnableOutput();
}
